// Generated by Cap'n Proto compiler, DO NOT EDIT
// source: c++.capnp

#pragma once

#include <capnp/generated-header-support.h>
#include <kj/windows-sanity.h>

#ifndef CAPNP_VERSION
#error "CAPNP_VERSION is not defined, is capnp/generated-header-support.h missing?"
#elif CAPNP_VERSION != 1004000
#error "Version mismatch between generated code and library headers.  You must use the same version of the Cap'n Proto compiler and library."
#endif


CAPNP_BEGIN_HEADER

namespace capnp {
namespace schemas {

CAPNP_DECLARE_SCHEMA(b9c6f99ebf805f2c);
CAPNP_DECLARE_SCHEMA(f264a779fef191ce);
CAPNP_DECLARE_SCHEMA(ac7096ff8cfc9dce);

}  // namespace schemas
}  // namespace capnp

namespace capnp {
namespace annotations {

// =======================================================================================

// =======================================================================================

}  // namespace
}  // namespace

CAPNP_END_HEADER

